#include <chrono>
#include <fstream>
#include <istream>
#include <ios>
#include <iostream>
#include <list>
#include <memory>
#include <string>
#include <sdsl/util.hpp>

#include "../../index/mti_lzss.hpp"

using namespace sdsl;
using namespace std;
using namespace std::chrono;
using timer = std::chrono::high_resolution_clock;

typedef typename mti_lzss::size_type size_type;

string to_latex_string(string s);
size_type findLZSSsize(const structure_tree_node *node);		

int main(int argc, char **argv) {
	if (argc < 4)	return 0; //nothing to do
	const char *tc_name = argv[1];
	const char *htmlsb_outfile = argv[2];

	//vars saving some measured times and spaces
	size_type acc_file_size = 0; //accumulated file sizes
	size_type c_time_ms = 0; //construction time in milliseconds
	size_type index_size = 0; //size of index in bytes
	size_type lzss_size = 0; //size of the lempel - ziv - parsing

	//// INDEX CREATION ///////////////////////////////////////////////////
	mti_lzss index;
	{
		//create a list containing streams to files
		list<unique_ptr<istream>> streamlist;
		argc -= 3; argv += 3;
		for (int arg = 0; arg < argc; arg++) {

			//open a ifstream for each file
			ifstream *s = new ifstream(argv[arg]);
			if (!s->good()) {
				cerr << "Unable to open file " << argv[arg] << endl;
				return 1;
			}
			streamlist.push_back(unique_ptr<istream>( s ));
			arg++;			
		}
		//build index using file stream list
		auto start = timer::now();
		index = mti_lzss( streamlist.begin(), streamlist.end() );
		auto stop = timer::now();
		c_time_ms = duration_cast<milliseconds>(stop-start).count();
	}

	//// REQUIRED INDEX SPACE /////////////////////////////////////////////
	
	//serialize index to get space requirements
	nullstream ns;
	structure_tree_node *stn = new structure_tree_node("name","");
	index_size = index.serialize(ns, stn, "name2");
	lzss_size  = findLZSSsize( stn );

	//write space breakdown
	{
		ofstream sb_stream(htmlsb_outfile, ios_base::trunc);
		if (!sb_stream.good()) {
			cerr << "Unable to write space breakdown to file " 
			     << htmlsb_outfile << endl;
			return 1;
		}
		write_structure<HTML_FORMAT>( index, sb_stream );
	}

	//// OUTPUT ALREADY MEASURED TIMES AND SPACES /////////////////////////
	cout << "\\providecommand{\\testcasename}{}" << endl;
	cout << "\\renewcommand{\\testcasename}{" << to_latex_string(tc_name) << "}" << endl;
	cout << "\\providecommand{\\testcasefiles}{}" << endl;
	cout << "\\renewcommand{\\testcasefiles}{" << endl;
	for (int arg = 0; arg < argc; arg++) {
		size_type filesize = util::file_size(argv[arg]);
		acc_file_size += filesize;

		cout << to_latex_string(util::basename(argv[arg])) << " & "
		     << filesize << " \\\\" << endl;
	}
	cout << "}" << endl;
	cout << "\\providecommand{\\tcaccfilesize}{}" << endl;
	cout << "\\renewcommand{\\tcaccfilesize}{" << acc_file_size << "}" << endl;
	cout << "\\providecommand{\\indexctime}{}" << endl;
	cout << "\\renewcommand{\\indexctime}{" << c_time_ms << "}" << endl;
	cout << "\\providecommand{\\indexsize}{}" << endl;
	cout << "\\renewcommand{\\indexsize}{" << index_size << "}" << endl;
	cout << "\\providecommand{\\lzsssize}{}" << endl;
	cout << "\\renewcommand{\\lzsssize}{" << lzss_size << "}" << endl;

	//// MEASURE PATTERN SEARCH TIME //////////////////////////////////////
	cout << "\\pgfplotstableread[row sep=\\\\]{" << endl;
	cout << "m occ time \\\\" << endl;
	string pattern = "";
	for (getline(cin, pattern, '\0'); !pattern.empty(); pattern = "", getline(cin, pattern, '\0')) {
		auto start = timer::now();
		auto occs = index.findPattern( pattern );
		auto stop = timer::now();
		size_type time = duration_cast<milliseconds>(stop-start).count();
		cout << pattern.size() << " " << occs.size() << " " << time << "\\\\" << endl;
	}
	cout << "}{\\patternsearchtimes}" << endl;
	return 0;
}

string to_latex_string(string s) {
	string r = "";
	for (unsigned int i = 0; i < s.size(); i++) {
		r += util::to_latex_string(s[i]);
	}
	return r;
}

size_type findLZSSsize(const structure_tree_node *node) {
	if (node) {
		if (node->name == "coding") {
			return node->size;
		}
		size_type size = 0;
		for (auto it = node->children.cbegin(); it != node->children.cend(); ++it) {
			size = max(size, findLZSSsize(it->second.get()));
		}
		return size;
	}
	return 0;
}
