#include "csi_lzss.hpp"

#include <assert.h>
#include <algorithm>
#include <functional>
#include <sdsl/bits.hpp>
#include <sdsl/construct_sa.hpp>
#include <sdsl/util.hpp>

typedef typename csi_lzss::size_type size_type;

//// CONSTRUCTION /////////////////////////////////////////////////////////////

void csi_lzss::construct(const unsigned char *text, const size_type t_len) {

	if (t_len == 0)	return;

	//construct suffix array from text
        sdsl::int_vector<> sa(t_len, 0, sdsl::bits::hi(t_len)+1);
        sdsl::algorithm::calculate_sa(text, t_len, sa);			

	//use suffix array to construct phrases
	coding = lzss_coding(text, t_len, sa);

	//set up a patricia trie for suffixes of phrases
	auto suffChar = [&](size_type p, size_type i) { return text[coding.phraseend(p) + i]; };
	auto suffLen  = [&](size_type p)              { return t_len - coding.phraseend(p); };
	pt_phrasesuffixes = patricia_trie(coding.phrasecount(), suffChar, suffLen, t_len);
	//and one for the reversed phrases
	auto revPChar = [&](size_type p, size_type i) { return text[coding.phraseend(p) - 1 - i]; };
	auto revPLen  = [&](size_type p)              { return coding.length(p); };
	pt_revphrases = patricia_trie(coding.phrasecount(), revPChar, revPLen, t_len);

	//set up revphrase -  suffix - grid

	//build a mapping from phrase index to lexicographic index in pt_phrasesuffixes,
	//so suffix lex order of a phrase can be associated with reversed phrase lex order.
	//Since sa is not used any more, use sa (covers enough space)
	sdsl::int_vector<> &phrase2sufflex = sa;
	for (size_type p = 0; p < coding.phrasecount(); p++) {
		phrase2sufflex[pt_phrasesuffixes.lex2str(p)] = p;
	}
	//build data for grid.
	sdsl::int_vector<> griddata(coding.phrasecount(), 0, 
	                            sdsl::bits::hi(coding.phrasecount()) + 1);
	for (size_type p = 0; p < coding.phrasecount(); p++) {
		griddata[p] = phrase2sufflex[pt_revphrases.lex2str(p)];
	}
	//finally construct grid
	construct_im(rs_grid, griddata);

	//build phrase source phrase - array for extracting text
	//use griddata since it is not used any more
	size_type non_value_phrasecount = 0;
	p_src_p = sdsl::int_vector<>(std::move(griddata));
	auto it = coding.begin(); //iterator over phrase ends of coding
	size_type p = 0; //actual phrases index
	while (it != coding.end()) {
		if (!coding.is_value(p)) { //only do something when phrase has a phrase source
			//use binary search to detect phrase which
			//contains phrase source of phrase p.
			auto low = std::upper_bound (coding.begin(), it, coding.source(p));
			p_src_p[p] = (low - coding.begin());

			++non_value_phrasecount;
		}
		++it; ++p;
	}

	//construct source - ordered phrase arrays
	p_ord_src    = sdsl::int_vector<>(non_value_phrasecount, 0, 
	                                  sdsl::bits::hi(coding.phrasecount()) + 1);
	//ensure array contains only phrase indices that contain a non - value
	size_type i = 0;
	for (p = 0; p < coding.phrasecount(); p++) {
		if (!coding.is_value(p)) {
			p_ord_src[i++] = p;
		}
	}
	p_ord_srcend = sdsl::int_vector<>(p_ord_src); //copy values into p_ord_srcend
	//now sort both
	auto phraseSrcCmp =    [&](size_type p1, size_type p2) 
	                       { return coding.source(p1) < coding.source(p2); };
	auto phraseSrcEndCmp = [&](size_type p1, size_type p2)
	                       { return coding.source(p1) + coding.length(p1)
	                              < coding.source(p2) + coding.length(p2); };
	std::sort(p_ord_src.begin(),    p_ord_src.end(),    phraseSrcCmp);
	std::sort(p_ord_srcend.begin(), p_ord_srcend.end(), phraseSrcEndCmp);

	//finally build grid for association
	//use same trick for grid setup as on previous grid
	sdsl::int_vector<> &phrase2encodedsrcend = sa;
	for (i = 0; i < p_ord_srcend.size(); i++) {
		phrase2encodedsrcend[p_ord_srcend[i]] = i;
	}
	griddata = sdsl::int_vector<>(p_ord_src.size(), 0, 
	                              sdsl::bits::hi(p_ord_src.size()) + 1);
	for (i = 0; i < p_ord_src.size(); i++) {
		griddata[i] = phrase2encodedsrcend[p_ord_src[i]];
	}
	construct_im(se_grid, griddata);
}

//// EXTRACTION ///////////////////////////////////////////////////////////////

bool csi_lzss::extract(size_type p, size_type skip, size_type count, 
             std::function<bool (unsigned char)> extractOn) const {
	//select correct phrase
	while (coding.length(p) <= skip) {
		skip -= coding.length(p);
		p++;
	}
	//extract count characters
	do {
		if (coding.is_value(p)) {
			if (!extractOn(coding.value(p)))
				return false;
			count--;
		}
		else {	//go to phrases source phrase
			size_type pp = p_src_p[p];
			size_type pskip = skip + coding.source(p) - coding.phrasestart(pp);
			size_type pcount = std::min(coding.length(p) - skip, count);
			if (!extract(pp, pskip, pcount, extractOn))
				return false;
			//push on vars
			count -= pcount;
			skip = 0;
		}
		p++;
	} while (count > 0);	
	return true;
}

void csi_lzss::extract(unsigned char *buff, size_type i, size_type m) const {
	assert(i < textlen() && i + m <= textlen());

	size_type pos = 0;
	auto saveChar = [&](unsigned char c) { buff[pos++] = c; return true; };

	if (i == 0) {
		extract(0, 0, m, saveChar); //easy case ;)
	} else {
		//use binary search to detect phrase where S[i..i+m-1] starts
		auto low = std::upper_bound(coding.begin(), coding.end(), i);
		size_type p = low - coding.begin();
		size_type p_start = coding.phrasestart(p);
			
		//and extract string
		extract(p, i - p_start, m, saveChar);
	}
}

//// FIND PATTERN /////////////////////////////////////////////////////////////

std::list<size_type> csi_lzss::findPattern(const unsigned char *p, size_type m) {
	std::list<size_type> occurences;
	if (m == 0 || m > textlen()) {
		return occurences;
	}

	//find primary occurences
	if (m == 1) { //special case to handle
		//find reversed phrases starting with character p[0]
		auto charAtRevP = [&](size_type) { return p[0]; };
		std::pair<size_type,size_type> range_rp = 
			pt_revphrases.partial_prefix_lex_range(charAtRevP, 1);
		
		//save all positions where the matching phrase contains
		//a value (all other occurences get done later by finding
		// secondary occurences)
		size_type i = range_rp.first;
		//This works because values are the lexicographic smallest strings
		while (i <= range_rp.second && coding.is_value(pt_revphrases.lex2str(i))) {
			occurences.push_back(coding.phrasestart(pt_revphrases.lex2str(i++)));
		}
	} else {
		for (size_type i = 1; i < m; i++) { //split pattern at position i

			//find lexicographic range of p[i-1 .. 0] in
			//reversed phrases trie
			auto charAtRevPi = [&](size_type j) { return p[i - 1 - j]; };
			std::pair<size_type,size_type> range_rp = 
				pt_revphrases.partial_prefix_lex_range(charAtRevPi, i);

			//and lexicographic range of p[i .. m] in
			//phrase suffix trie
			auto charAtSuffPi = [&](size_type j) { return p[i + j]; };
			std::pair<size_type,size_type> range_ps =
				pt_phrasesuffixes.partial_prefix_lex_range(charAtSuffPi, m - i);
				
			//found association between reversed phrases and suffixes using grid
			auto range = rs_grid.range_search_2d(range_rp.first, range_rp.second, 
			                                  range_ps.first, range_ps.second);
			auto &points = range.second;
			//check if points were found
			if (!points.empty()) {
				//by now, we only partially match s against reversed phrases and suffixes,
				// so do exactly 1 match to be sure points contains substrings
				size_type phrase = pt_revphrases.lex2str(points[0].first);
				size_type phrase_pos = 0; //actual position in pattern
				auto pCharMatches = [&](unsigned char c) { return p[phrase_pos++] == c; };
				if (extract(phrase, coding.length(phrase) - i, m, pCharMatches)) {
					
					//finally, we found primary occurences, so put them to list
					for (auto &point : points) {
						phrase = pt_revphrases.lex2str(point.first);
						occurences.push_back( coding.phraseend(phrase) - i );
					}
				}
			}
		}
	}

	//find secondary occurences
	auto it = occurences.begin();
	while (it != occurences.end()) {
		size_type pos = *it++;
		//compute count of phrases with source(p) <= pos
		auto sourceCmp = [&](size_type startpos, size_type p) 
		                    { return startpos < coding.source(p); };
		size_type src_before = std::upper_bound(p_ord_src.begin(), 
		                                        p_ord_src.end(), 
		                                        pos, sourceCmp )
		                     - p_ord_src.begin();

		if (src_before > 0) { //ensure at least 1 phrase has a source before pos
			//compute count of phrases with source(p) + length(p) < pos + m
			auto sourceEndCmp = [&](size_type p, size_type endpos)
			                       { return coding.source(p) + coding.length(p) < endpos; };
			size_type srcend_before = std::lower_bound(p_ord_srcend.begin(),
			                                           p_ord_srcend.end(),
			                                           pos + m, sourceEndCmp )
			                        - p_ord_srcend.begin();

			if (srcend_before < p_ord_srcend.size()) { //ensure at least 1 phrase
				//has a source end after occurence end position

				//now, use range search of grid to detect all phrases with
				// source(p) <= pos and source(p) + length(p) >= pos + m,
				//what means that p is a secondary occurence
				auto range = se_grid.range_search_2d(
				             0, src_before - 1, srcend_before, p_ord_srcend.size() - 1);

				auto &points = range.second;
				//insert points before actual iterator, and set iterator onto first
				//inserted value
				for (auto &point : points) {
					size_type p = p_ord_src[point.first];
					
					size_type occ_pos = coding.phrasestart(p) + (pos - coding.source(p));
					it = occurences.insert(it, occ_pos);
				}
			}
		}
	}
	return occurences;
}

//// SERIALIZATION AND LOAD ///////////////////////////////////////////////////

size_type csi_lzss::serialize(std::ostream& out, sdsl::structure_tree_node* v,
		std::string name) const {
	sdsl::structure_tree_node *child = sdsl::structure_tree::add_child(
		                                     v, name, sdsl::util::class_name(*this));
	size_type written_bytes = 0;
	written_bytes += coding.serialize(out, child, "coding");
	written_bytes += p_src_p.serialize(out, child, "p_src_p");
	written_bytes += pt_revphrases.serialize(out, child, "pt_revphrases");
	written_bytes += pt_phrasesuffixes.serialize(out, child, "pt_phrasesuffixes");
	written_bytes += rs_grid.serialize(out, child, "rs_grid");
	written_bytes += p_ord_src.serialize(out, child, "p_ord_src");
	written_bytes += p_ord_srcend.serialize(out, child, "p_ord_srcend");
	written_bytes += se_grid.serialize(out, child, "se_grid");
	sdsl::structure_tree::add_size(child, written_bytes);
	return written_bytes;
}

void csi_lzss::load(std::istream& in) {
	coding.load(in);
	p_src_p.load(in);
	pt_revphrases.load(in);
	pt_phrasesuffixes.load(in);
	rs_grid.load(in);
	p_ord_src.load(in);
	p_ord_srcend.load(in);
	se_grid.load(in);
}
