#ifndef __SI_LZSS_CODING_H
#define __SI_LZSS_CODING_H

#include <assert.h>
#include <sdsl/int_vector.hpp>
#include <sdsl/io.hpp>

//! A Variant of Lempel - Ziv- Storer - Szymanski Coding
/*!	\author Uwe Baier
 *
 *	Beside the normal lzss coding, this class provides extra information
 *	that covers the position of the phrase's end and start as 
 *	absolute position in text (with same space requirements as normal lzss coding).
 *
 *	Instead of the normal (source, length) - pairs of lzss, this variant saves pairs
 *	of (source, end) where end is the absolute (exclusive) position in text where 
 *	phrase ends. To compute length, one may use that 
 *	 length(p) = end[p] - end[p-1] for p >= 1 and length(0) = 1.
 *	To detect if a phrase p contains a value, simply check if length(p) = 1
 *	(value can be fetched from source[p]).
 *
 *	By this fact, the coding may contains multiple equal values, but that doesn't
 *	affect the needed space for the coding, which is 2 * #phrasecount * log(n).
 *
 *	Factorization algorithm uses suffix array and works linear.
 *
 * 	\par Space complexity
 *      	\f$ 2 n' \log n \f$
 *      	where n is length of text and n' is count of lzss - phrases.
 *
 *	\par References
 *      [1] E. Ohlebusch:
 *          ,,Lempel-Ziv Factorization: LZ77 without Window'',
 *          June 2, 2014.
 *	[2] J. Ziv and A. Lempel:
 *	    ,,A universal algorithm for sequential data compression'',
 *	    IEEE Transactions on Information Theory 23 (1977) 337 - 343.
 */
class lzss_coding {
	public:
		typedef typename sdsl::int_vector<>::size_type	size_type;
	private:
		sdsl::int_vector<> pend;    //end of phrases (absolute position in text)
		sdsl::int_vector<> src;     //source of phrase

		//utility for construction
		size_type parse_phrase(const unsigned char *text, 
		                       const size_type t_len, size_type pos,
		                       size_type psv_text, size_type nsv_text,
		                       size_type phrase);
	public:
		//! Default constructor
		lzss_coding();

		//! Constructs a Lempel - Ziv- Storer - Szymanski Coding from a text
		/*! \param text  the original text to be encoded. 
		                 text[t_len] = 0 and for i<len text[i] != 0.
		    \param t_len length of the text
		    \param sa    suffix array of the text.
		*/	
		lzss_coding(const unsigned char *text, const size_type t_len, 
		            const sdsl::int_vector<> &sa);

		//! Returns if this coding is empty
		bool empty() const {
			return pend.empty();
		}

		//! Returns the count of phrases in this coding.
		size_type phrasecount() const {
			return pend.size();
		}

		//! Returns the length of phrase p.
		size_type length(size_type p) const {
			assert(p < pend.size());
			return (p == 0) ? 1 : pend[p] - pend[p-1];
		}

		//! Returns whether phrase p contains a value.
		bool is_value(size_type p) const {
			return length(p) == 1;
		}

		//! Returns the value of phrase p (if it contains one).
		unsigned char value(size_type p) const {
			assert(is_value(p));
			return (unsigned char)src[p];
		}

		//! Returns the source of phrase p(if it contains no value).
		size_type source(size_type p) const {
			assert(!is_value(p));
			return src[p];
		}

		//! Returns the absolute position in text of p's start
		// (e.g. phrasestart(0) = 0)
		size_type phrasestart(size_type p) const {
			assert(p < pend.size());
			return (p == 0) ? 0 : pend[p-1];
		}

		//! Returns the absolute position in text of p's exclusive end
		//  (e.g. phraseend(phrasecount() - 1) = length of text)
		size_type phraseend(size_type p) const {
			assert(p < pend.size());
			return pend[p];
		}

		typedef typename sdsl::int_vector<>::const_iterator const_iterator;

		//! Returns a start iterator what iterates over the phrase ends 
		//  of this coding's phrases.
		const_iterator begin() const {
			return pend.begin();
		}

		//! Returns a start iterator what iterates over the phrase ends 
		//  of this coding's phrases.
		const_iterator end() const {
			return pend.end();
		}	

        	//! Serializes the data structure into the given ostream
        	size_type serialize(std::ostream& out, sdsl::structure_tree_node* v=nullptr,
                            std::string name="") const;

		//! Loads the data structure from the given istream.
		void load(std::istream& in);
};

#endif
