#include "patricia_trie.hpp"

#include <assert.h>
#include <algorithm>
#include <sdsl/bits.hpp>
#include <sdsl/util.hpp>

typedef typename patricia_trie::size_type size_type;
typedef typename patricia_trie::value_type value_type;

//// CONSTRUCTION /////////////////////////////////////////////////////////////

// creates a node and returns it's new index
size_type patricia_trie::createNode(size_type _lm_leaf, size_type _rm_leaf, 
		size_type _e_lbl_len, size_type &nodecount) {
	lm_leaf[nodecount] = _lm_leaf;
	rm_leaf[nodecount] = _rm_leaf;
	e_lbl_len[nodecount] = _e_lbl_len;
	children.emplace_back();
	return nodecount++;
}

void patricia_trie::update_lm_leaf(std::stack<size_type> &&nodestack, 
		size_type old_lm_leaf_val, size_type new_lm_leaf_val) {
	while (!nodestack.empty() && lm_leaf[nodestack.top()] == old_lm_leaf_val) {
		lm_leaf[nodestack.top()] = new_lm_leaf_val;
		nodestack.pop();
	}
}

void patricia_trie::update_rm_leaf(std::stack<size_type> &&nodestack,
		size_type old_rm_leaf_val, size_type new_rm_leaf_val) {
	while (!nodestack.empty() && rm_leaf[nodestack.top()] == old_rm_leaf_val) {
		rm_leaf[nodestack.top()] = new_rm_leaf_val;
		nodestack.pop();
	}
}

void patricia_trie::insert(value_type i, 
		std::function<unsigned char (value_type,size_type)> strChar,
		std::function<size_type (value_type)> strLen,
		sdsl::int_vector<> &nextleaf, size_type &nodecount) {

	//create a leaf
	size_type l_idx = i; //index of leaf
	nextleaf[i] = l_idx;
	
	//start iterating the tree
	size_type strpos = 0;
	size_type n = 0; //actual node (starting at root)
	std::stack<size_type> nodestack; //a stack saving the traversal of nodes
	while (strpos < strLen(i)) {
		nodestack.push(n);
		unsigned char c = strChar(i,strpos);
		auto it = children[n].find(c);
		if (it == children[n].end()) { //character was not used in node's children yet
			//create a node pointing onto leaf
			size_type n_idx = createNode(l_idx, l_idx, strLen(i) - strpos, nodecount);

			//let node n point to new node
			auto p = children[n].insert(std::pair<unsigned char,size_type>(c, n_idx));
			it = p.first;

			//update leftmost / rightmost leaf "pointers"
			if (it == --children[n].end()) {
				update_rm_leaf(std::move(nodestack), rm_leaf[n], l_idx);
			}
			//ensure lm_leaf is lexicographic greater than string
			else if (it == children[n].begin() &&
			         strLen(lm_leaf[n]) != strpos) {
				update_lm_leaf(std::move(nodestack), lm_leaf[n], l_idx);
			}
			
			return;
		} else { //character is used in node
			n = it->second; //place n to next node

			//check if edge label matches
			for (size_type j = 1; j < e_lbl_len[n]; j++) {

				unsigned char c2 = strChar(lm_leaf[n], strpos + j);
				if (strpos + j == strLen(i) || strChar(i, strpos + j) != c2) {
					//edge label does not match on string, 
					//so create a node on current edge
					size_type nn_idx = createNode(lm_leaf[n], rm_leaf[n], j, nodecount);
					//lm_leaf and rm_leaf get updated on next iteration or end of loop
		
					//update rest what is necessary
					children[nn_idx][c2] = n;
					e_lbl_len[n] -= j;
					children[nodestack.top()][c] = nn_idx;

					n = nn_idx;
					break; //to avoid j++
				}
			}
			strpos += e_lbl_len[n];		
		}
	}
	//by here, we stand on a node where string of leftmost leaf
	//string is less or equal to this one's, so update lm_leaf of actual node
	//swap nextleaf[l_idx] and lm_leaf[n]
	size_type temp = nextleaf[l_idx];
	nextleaf[l_idx] = lm_leaf[n];
	lm_leaf[n] = temp;
	update_lm_leaf(std::move(nodestack), nextleaf[l_idx], lm_leaf[n]);
}

void patricia_trie::sortLeafsLex(size_type n, size_type nodetextlen, size_type &l,
		std::function<size_type (value_type)> strLen,
		const sdsl::int_vector<> &nextleaf) {

	//l is used for the position where the next leaf will be inserted

	nodetextlen += e_lbl_len[n];
	size_type l_idx = lm_leaf[n];
	//update lm_leaf
	lm_leaf[n] = l;

	//the lexicographic smallest strings are those, who have same length as
	//actual node text len, so store them as first
	while (strLen(l_idx) == nodetextlen) {
		leafs[l++] = l_idx;

		if (l_idx == nextleaf[l_idx]) break; //end of leaf list of this node
		l_idx = nextleaf[l_idx];
	}
	//get recursive on all children
	for (auto &child : children[n]) {
		sortLeafsLex(child.second, nodetextlen, l, strLen, nextleaf);
	}
	//and update rm_leaf
	rm_leaf[n] = l - 1;	
}

void patricia_trie::createEmptyTrie() {
	//empty range of [lm_leaf..rm_leaf] ([1,0])
	lm_leaf = sdsl::int_vector<>(1, 1, 1);
	rm_leaf = sdsl::int_vector<>(1, 0, 1);
	
	e_lbl_len = sdsl::int_vector<>(1, 0, 1);
	children.emplace_back();
}

patricia_trie::patricia_trie(size_type n,
		std::function<unsigned char (value_type,size_type)> strChar,
		std::function<size_type (value_type)> strLen,
		size_type maxstrlen) {
	
	//check trie not to be empty
	if (n == 0) {
		createEmptyTrie(); 
		return;
	}
	//if maximum string length is not given, determine it
	if (maxstrlen == 0) {
		for (value_type i = 0; i < n; i++) {
			maxstrlen = std::max(maxstrlen, strLen(i));
		}
	}
	//set up all needed integer vectors
	lm_leaf = sdsl::int_vector<>(2*n, 0, sdsl::bits::hi(n) + 1);
	rm_leaf = sdsl::int_vector<>(2*n, 0, sdsl::bits::hi(n) + 1);
	e_lbl_len = sdsl::int_vector<>(2*n, 0, sdsl::bits::hi(maxstrlen) + 1);
	children.reserve(2 * n);
	leafs = sdsl::int_vector<>(n, 0, sdsl::bits::hi(n) + 1);

	//create root node by doing nothing ;) (lm_leaf = rm_leaf = e_lbl_len = 0)
	children.emplace_back(); //create first map to avoid crash
	size_type nodecount = 1;

	//build temporary structures needed for trie construction
	//needed for duplicate entries, list of strings that are duplicate
	sdsl::int_vector<> nextleaf(n, 0, sdsl::bits::hi(n) + 1);

	//put elements in temporary trie
	for (value_type i = 0; i < n; i++) {
		insert(i, strChar, strLen, nextleaf, nodecount);
	}

	//by now, elements are in trie, but the leafs are not sorted lexicographically
	// (means templeaf[i] need not to be less or equal to templeaf[i+1])
	// so iterate tree to sort them
	size_type l = 0;
	sortLeafsLex(0, 0, l, strLen, nextleaf);

	//and finally trim everything to correct size
	lm_leaf.resize(nodecount);
	rm_leaf.resize(nodecount);
	e_lbl_len.resize(nodecount);
	children.resize(nodecount);
}

patricia_trie::patricia_trie() {
	createEmptyTrie();
}

//// LEX RANGES ///////////////////////////////////////////////////////////////

std::pair<size_type,size_type> patricia_trie::partial_prefix_lex_range(
		std::function<unsigned char (size_type)> charAtPrefix, 
		size_type prefixlen) const {
	size_type n = 0; //actual node
	size_type strpos = 0;
	while (strpos < prefixlen) {
		auto it = children[n].find( charAtPrefix(strpos) );
		if (it == children[n].end()) { //character not found
			return std::pair<size_type,size_type>(1,0); //return empty range
		}
		n = it->second;
		strpos += e_lbl_len[n];
	}
	return std::pair<size_type,size_type>(lm_leaf[n], rm_leaf[n]);
}

std::pair<size_type,size_type> patricia_trie::prefix_lex_range(
		std::function<unsigned char (value_type,size_type)> strChar,
		std::function<unsigned char (size_type)> charAtPrefix, 
		size_type prefixlen) const {
	size_type n = 0; //actual node
	size_type strpos = 0;
	while (strpos < prefixlen) {
		auto it = children[n].find( charAtPrefix(strpos) );
		if (it == children[n].end()) { //character not found
			return std::pair<size_type,size_type>(1, 0); //return empty range
		}
		//go to node
		n = it->second;
		//and compare edge labels
		value_type n_lm_leaf_str = leafs[lm_leaf[n]];
		for (size_type j = 1; j < e_lbl_len[n] && strpos + j < prefixlen; j++) {
			if (charAtPrefix(strpos + j) != strChar(n_lm_leaf_str, strpos + j)) {
				//characters not equal, so stop
				return std::pair<size_type,size_type>(1,0);
			}
		}
		strpos += e_lbl_len[n];
	}
	return std::pair<size_type,size_type>(lm_leaf[n], rm_leaf[n]);
}

//// SERIALIZATION AND LOAD ///////////////////////////////////////////////////

size_type patricia_trie::serialize(std::ostream& out, sdsl::structure_tree_node* v,
		std::string name) const {
	sdsl::structure_tree_node *child = sdsl::structure_tree::add_child(
		                                     v, name, sdsl::util::class_name(*this));
	size_type written_bytes = 0;
	written_bytes += lm_leaf.serialize(out, child, "lm_leaf");
	written_bytes += rm_leaf.serialize(out, child, "rm_leaf");
	written_bytes += e_lbl_len.serialize(out, child, "e_lbl_len");
	//serialize vector of maps
	size_type m_bytes = 0;
	sdsl::structure_tree_node *m_child = sdsl::structure_tree::add_child(
		                                     child, "children", sdsl::util::class_name(children));
	for (auto &m : children) {
		//save length of map
		out.put((uint8_t)m.size()); //must be < 255 since '0' is not contained
		m_bytes++;

		//and their entries
		for (auto &entry : m) {
			out.write((char *)&entry.first, sizeof(unsigned char));
			out.write((char *)&entry.second, sizeof(size_type));
			m_bytes += sizeof(unsigned char) + sizeof(size_type);
		}
	}
	sdsl::structure_tree::add_size(m_child, m_bytes);
	written_bytes += m_bytes;

	written_bytes += leafs.serialize(out, child, "leafs");
	sdsl::structure_tree::add_size(child, written_bytes);
	return written_bytes;
}

void patricia_trie::load(std::istream& in) {
	lm_leaf.load(in);
	rm_leaf.load(in);
	e_lbl_len.load(in);
	//load vector map
	children = std::vector<std::map<unsigned char,size_type>>(lm_leaf.size());
	for (auto &m : children) {
		//read how much entries must be loaded
		uint8_t entrycount = (uint8_t)in.get();
		while (entrycount--) {
			std::pair<unsigned char, size_type> entry;
			in.read((char *)&entry.first, sizeof(unsigned char));
			in.read((char *)&entry.second, sizeof(size_type));
			m.insert( entry );
		}
	} 
	leafs.load(in);
}
